@@ -1,5 +1,5 @@
 /* snac - A simple, minimalistic ActivityPub instance */
-/* copyright (c) 2022 - 2023 grunfink et al. / MIT license */
+/* copyright (c) 2022 - 2024 grunfink et al. / MIT license */
 
 #include "xs.h"
 #include "xs_json.h"
@@ -60,18 +60,22 @@ const char *default_avatar_base64(void)
 }
 
 
-int activitypub_request(snac *snac, const char *url, xs_dict **data)
+int activitypub_request(snac *user, const char *url, xs_dict **data)
 /* request an object */
 {
-    int status;
+    int status = 0;
     xs *response = NULL;
     xs *payload = NULL;
     int p_size;
     char *ctype;
 
-    /* get from the net */
-    response = http_signed_request(snac, "GET", url,
-        NULL, NULL, 0, &status, &payload, &p_size, 0);
+    *data = NULL;
+
+    if (user != NULL) {
+        /* get from the net */
+        response = http_signed_request(user, "GET", url,
+            NULL, NULL, 0, &status, &payload, &p_size, 0);
+    }
 
     if (status == 0 || (status >= 500 && status <= 599)) {
         /* I found an instance running Misskey that returned
@@ -95,23 +99,26 @@ int activitypub_request(snac *snac, const char *url, xs_dict **data)
             status = 400;
         else
         if (xs_str_in(ctype, "application/activity+json") != -1 ||
-            xs_str_in(ctype, "application/ld+json") != -1)
-            *data = xs_json_loads(payload);
+            xs_str_in(ctype, "application/ld+json") != -1) {
+
+            /* if there is no payload, fail */
+            if (xs_is_null(payload))
+                status = 400;
+            else
+                *data = xs_json_loads(payload);
+        }
         else
             status = 500;
     }
 
-    if (!valid_status(status))
-        *data = NULL;
-
     return status;
 }
 
 
-int actor_request(snac *snac, const char *actor, xs_dict **data)
+int actor_request(snac *user, const char *actor, xs_dict **data)
 /* request an actor */
 {
-    int status, status2;
+    int status;
     xs *payload = NULL;
 
     if (data)
@@ -122,9 +129,9 @@ int actor_request(snac *snac, const char *actor, xs_dict **data)
 
     if (status != 200) {
         /* actor data non-existent or stale: get from the net */
-        status2 = activitypub_request(snac, actor, &payload);
+        status = activitypub_request(user, actor, &payload);
 
-        if (valid_status(status2)) {
+        if (valid_status(status)) {
             /* renew data */
             status = actor_add(actor, payload);
 
@@ -133,6 +140,8 @@ int actor_request(snac *snac, const char *actor, xs_dict **data)
                 payload = NULL;
             }
         }
+        else
+            srv_debug(1, xs_fmt("actor_request error %s %d", actor, status));
     }
 
     /* collect the (presumed) shared inbox in this actor */
@@ -181,7 +190,7 @@ int timeline_request(snac *snac, char **id, xs_str **wrk, int level)
                 if (xs_is_null(type))
                     type = "(null)";
 
-                srv_debug(1, xs_fmt("timeline_request type %s '%s'", nid, type));
+                srv_debug(2, xs_fmt("timeline_request type %s '%s'", nid, type));
 
                 if (strcmp(type, "Create") == 0) {
                     /* some software like lemmy nest Announce + Create + Note */
@@ -316,13 +325,13 @@ int send_to_inbox(snac *snac, const xs_str *inbox, const xs_dict *msg,
 }
 
 
-xs_str *get_actor_inbox(snac *snac, const char *actor)
+xs_str *get_actor_inbox(const char *actor)
 /* gets an actor's inbox */
 {
     xs *data = NULL;
     char *v = NULL;
 
-    if (valid_status(actor_request(snac, actor, &data))) {
+    if (valid_status(actor_request(NULL, actor, &data))) {
         /* try first endpoints/sharedInbox */
         if ((v = xs_dict_get(data, "endpoints")))
             v = xs_dict_get(v, "sharedInbox");
@@ -341,7 +350,7 @@ int send_to_actor(snac *snac, const char *actor, const xs_dict *msg,
 /* sends a message to an actor */
 {
     int status = 400;
-    xs *inbox = get_actor_inbox(snac, actor);
+    xs *inbox = get_actor_inbox(actor);
 
     if (!xs_is_null(inbox))
         status = send_to_inbox(snac, inbox, msg, payload, p_size, timeout);
@@ -438,7 +447,8 @@ int is_msg_public(const xs_dict *msg)
 int is_msg_for_me(snac *snac, const xs_dict *c_msg)
 /* checks if this message is for me */
 {
-    const char *type = xs_dict_get(c_msg, "type");
+    const char *type  = xs_dict_get(c_msg, "type");
+    const char *actor = xs_dict_get(c_msg, "actor");
 
     if (xs_match(type, "Like|Announce")) {
         const char *object = xs_dict_get(c_msg, "object");
@@ -455,12 +465,34 @@ int is_msg_for_me(snac *snac, const xs_dict *c_msg)
             return 2;
 
         /* if it's by someone we don't follow, reject */
-        if (!following_check(snac, xs_dict_get(c_msg, "actor")))
-            return 0;
+        return following_check(snac, actor);
     }
 
-    /* if it's not a Create, allow */
-    if (strcmp(type, "Create") != 0)
+    /* if it's an Undo, it must be from someone related to us */
+    if (xs_match(type, "Undo")) {
+        return follower_check(snac, actor) || following_check(snac, actor);
+    }
+
+    /* if it's an Accept + Follow, it must be for a Follow we created */
+    if (xs_match(type, "Accept")) {
+        return following_check(snac, actor);
+    }
+
+    /* if it's a Follow, it must be explicitly for us */
+    if (xs_match(type, "Follow")) {
+        char *object = xs_dict_get(c_msg, "object");
+        return !xs_is_null(object) && strcmp(snac->actor, object) == 0;
+    }
+
+    /* if it's not a Create or Update, allow as is */
+    if (!xs_match(type, "Create|Update")) {
+        return 1;
+    }
+
+    int pub_msg = is_msg_public(c_msg);
+
+    /* if this message is public and we follow the actor of this post, allow */
+    if (pub_msg && following_check(snac, actor))
         return 1;
 
     xs_dict *msg = xs_dict_get(c_msg, "object");
@@ -468,20 +500,40 @@ int is_msg_for_me(snac *snac, const xs_dict *c_msg)
     xs_list *p = rcpts;
     xs_str *v;
 
+    xs *actor_followers = NULL;
+
+    if (!pub_msg) {
+        /* not a public message; get the actor and its followers list */
+        xs *actor_obj = NULL;
+
+        if (valid_status(object_get(actor, &actor_obj))) {
+            if ((v = xs_dict_get(actor_obj, "followers")))
+                actor_followers = xs_dup(v);
+        }
+    }
+
     while(xs_list_iter(&p, &v)) {
         /* explicitly for me? accept */
         if (strcmp(v, snac->actor) == 0)
             return 2;
 
-        /* for someone we follow? (probably cc'ed) accept */
-        if (following_check(snac, v))
-            return 5;
+        if (pub_msg) {
+            /* a public message for someone we follow? (probably cc'ed) accept */
+            if (following_check(snac, v))
+                return 5;
+        }
+        else
+        if (actor_followers && strcmp(v, actor_followers) == 0) {
+            /* if this message is for this actor's followers, are we one of them? */
+            if (following_check(snac, actor))
+                return 6;
+        }
     }
 
     /* accept if it's by someone we follow */
     char *atto = xs_dict_get(msg, "attributedTo");
 
-    if (!xs_is_null(atto) && following_check(snac, atto))
+    if (pub_msg && !xs_is_null(atto) && following_check(snac, atto))
         return 3;
 
     /* is this message a reply to another? */
@@ -494,7 +546,7 @@ int is_msg_for_me(snac *snac, const xs_dict *c_msg)
             atto = xs_dict_get(r_msg, "attributedTo");
 
             /* accept if the replied message is from someone we follow */
-            if (!xs_is_null(atto) && following_check(snac, atto))
+            if (pub_msg && !xs_is_null(atto) && following_check(snac, atto))
                 return 4;
         }
     }
@@ -1002,6 +1054,14 @@ xs_dict *msg_actor(snac *snac)
         msg = xs_dict_set(msg, "attachment", attach);
     }
 
+    /* use shared inboxes? */
+    if (xs_type(xs_dict_get(srv_config, "shared_inboxes")) == XSTYPE_TRUE) {
+        xs *d = xs_dict_new();
+        xs *si = xs_fmt("%s/shared-inbox", srv_baseurl);
+        d = xs_dict_append(d, "sharedInbox", si);
+        msg = xs_dict_set(msg, "endpoints", d);
+    }
+
     return msg;
 }
 
@@ -1453,27 +1513,28 @@ int update_question(snac *user, const char *id)
 
 int process_input_message(snac *snac, xs_dict *msg, xs_dict *req)
 /* processes an ActivityPub message from the input queue */
+/* return values: -1, fatal error; 0, transient error, retry;
+   1, processed and done; 2, propagate to users (only when no user is set) */
 {
-    /* actor and type exist, were checked previously */
-    char *actor  = xs_dict_get(msg, "actor");
-    char *type   = xs_dict_get(msg, "type");
+    char *actor = xs_dict_get(msg, "actor");
+    char *type  = xs_dict_get(msg, "type");
     xs *actor_o = NULL;
     int a_status;
     int do_notify = 0;
 
+    if (xs_is_null(actor) || *actor == '\0') {
+        srv_debug(0, xs_fmt("malformed message (bad actor)"));
+        return -1;
+    }
+
     /* question votes may not have a type */
     if (xs_is_null(type))
         type = "Note";
 
     /* reject uninteresting messages right now */
     if (strcmp(type, "Add") == 0) {
-        snac_debug(snac, 0, xs_fmt("Ignored message of type '%s'", type));
-        return 1;
-    }
-
-    if (xs_is_null(actor)) {
-        snac_debug(snac, 0, xs_fmt("malformed message"));
-        return 1;
+        srv_debug(0, xs_fmt("Ignored message of type '%s'", type));
+        return -1;
     }
 
     char *object, *utype;
@@ -1484,6 +1545,66 @@ int process_input_message(snac *snac, xs_dict *msg, xs_dict *req)
     else
         utype = "(null)";
 
+    /* special case for Delete messages */
+    if (strcmp(type, "Delete") == 0) {
+        /* if the actor is not here, do not even try */
+        if (!object_here(actor)) {
+            srv_debug(1, xs_fmt("dropped 'Delete' message from unknown actor '%s'", actor));
+            return -1;
+        }
+
+        /* discard crap */
+        if (xs_is_null(object)) {
+            srv_log(xs_fmt("dropped 'Delete' message with invalid object from actor '%s'", actor));
+            return -1;
+        }
+
+        /* also discard if the object to be deleted is not here */
+        char *obj_id = object;
+        if (xs_type(obj_id) == XSTYPE_DICT)
+            obj_id = xs_dict_get(obj_id, "id");
+
+        if (!object_here(obj_id)) {
+            srv_debug(1, xs_fmt("dropped 'Delete' message from unknown object '%s'", obj_id));
+            return -1;
+        }
+    }
+
+    /* bring the actor */
+    a_status = actor_request(snac, actor, &actor_o);
+
+    /* do not retry permanent failures */
+    if (a_status == 404 || a_status == 410 || a_status < 0) {
+        srv_debug(1, xs_fmt("dropping message due to actor error %s %d", actor, a_status));
+        return -1;
+    }
+
+    if (!valid_status(a_status)) {
+        /* do not retry 'Delete' messages */
+        if (strcmp(type, "Delete") == 0) {
+            srv_debug(1, xs_fmt("dropping 'Delete' message due to actor error %s %d", actor, a_status));
+            return -1;
+        }
+
+        /* other actor download errors may need a retry */
+        srv_debug(1, xs_fmt("error requesting actor %s %d -- retry later", actor, a_status));
+        return 0;
+    }
+
+    /* check the signature */
+    xs *sig_err = NULL;
+
+    if (!check_signature(req, &sig_err)) {
+        srv_log(xs_fmt("bad signature %s (%s)", actor, sig_err));
+
+        srv_archive_error("check_signature", sig_err, req, msg);
+        return -1;
+    }
+
+    /* if no user is set, no further checks can be done; propagate */
+    if (snac == NULL)
+        return 2;
+
     /* reject messages that are not for this user */
     if (!is_msg_for_me(snac, msg)) {
         snac_debug(snac, 1, xs_fmt("message from %s of type '%s' not for us", actor, type));
@@ -1501,38 +1622,15 @@ int process_input_message(snac *snac, xs_dict *msg, xs_dict *req)
         }
     }
 
-    /* bring the actor */
-    a_status = actor_request(snac, actor, &actor_o);
-
-    /* do not retry permanent failures */
-    if (a_status == 404 || a_status == 410 || a_status < 0) {
-        snac_debug(snac, 1,
-            xs_fmt("dropping message due to actor error %s %d", actor, a_status));
-
-        return 1;
-    }
-
-    if (!valid_status(a_status)) {
-        /* other actor download errors may need a retry */
-        snac_debug(snac, 1,
-            xs_fmt("error requesting actor %s %d -- retry later", actor, a_status));
-
-        return 0;
-    }
-
-    /* check the signature */
-    xs *sig_err = NULL;
-
-    if (!check_signature(snac, req, &sig_err)) {
-        snac_log(snac, xs_fmt("bad signature %s (%s)", actor, sig_err));
-
-        srv_archive_error("check_signature", sig_err, req, msg);
-
-        return 1;
-    }
-
     if (strcmp(type, "Follow") == 0) { /** **/
         if (!follower_check(snac, actor)) {
+            /* ensure the actor object is here */
+            if (!object_here(actor)) {
+                xs *actor_obj = NULL;
+                actor_request(snac, actor, &actor_obj);
+                object_add(actor, actor_obj);
+            }
+
             xs *f_msg = xs_dup(msg);
             xs *reply = msg_accept(snac, f_msg, actor);
 
@@ -1556,6 +1654,9 @@ int process_input_message(snac *snac, xs_dict *msg, xs_dict *req)
     }
     else
     if (strcmp(type, "Undo") == 0) { /** **/
+        if (xs_type(object) != XSTYPE_DICT)
+            utype = "Follow";
+
         if (strcmp(utype, "Follow") == 0) { /** **/
             if (valid_status(follower_del(snac, actor))) {
                 snac_log(snac, xs_fmt("no longer following us %s", actor));
@@ -1569,8 +1670,10 @@ int process_input_message(snac *snac, xs_dict *msg, xs_dict *req)
     }
     else
     if (strcmp(type, "Create") == 0) { /** **/
-        if (is_muted(snac, actor))
+        if (is_muted(snac, actor)) {
             snac_log(snac, xs_fmt("ignored 'Create' + '%s' from muted actor %s", utype, actor));
+            return 1;
+        }
 
         if (strcmp(utype, "Note") == 0) { /** **/
             char *id          = xs_dict_get(object, "id");
@@ -1650,6 +1753,9 @@ int process_input_message(snac *snac, xs_dict *msg, xs_dict *req)
         if (xs_type(object) == XSTYPE_DICT)
             object = xs_dict_get(object, "id");
 
+        if (is_muted(snac, actor) && !xs_startswith(object, snac->actor))
+            snac_log(snac, xs_fmt("dropped 'Announce' from muted actor %s", actor));
+        else
         if (is_limited(snac, actor) && !xs_startswith(object, snac->actor))
             snac_log(snac, xs_fmt("dropped 'Announce' from limited actor %s", actor));
         else {
@@ -1671,13 +1777,13 @@ int process_input_message(snac *snac, xs_dict *msg, xs_dict *req)
                         do_notify = 1;
                     }
                     else
-                        snac_log(snac, xs_fmt("dropped 'Announce' on actor request error %s", who));
+                        snac_debug(snac, 1, xs_fmt("dropped 'Announce' on actor request error %s", who));
                 }
                 else
                     snac_log(snac, xs_fmt("ignored 'Announce' about muted actor %s", who));
             }
             else
-                snac_debug(snac, 1, xs_fmt("error requesting 'Announce' object %s", object));
+                snac_debug(snac, 2, xs_fmt("error requesting 'Announce' object %s", object));
         }
     }
     else
@@ -1722,8 +1828,10 @@ int process_input_message(snac *snac, xs_dict *msg, xs_dict *req)
         if (xs_type(object) == XSTYPE_DICT)
             object = xs_dict_get(object, "id");
 
-        if (valid_status(timeline_del(snac, object)))
+        if (object_here(object)) {
+            timeline_del(snac, object);
             snac_debug(snac, 1, xs_fmt("new 'Delete' %s %s", actor, object));
+        }
         else
             snac_debug(snac, 1, xs_fmt("ignored 'Delete' for unknown object %s", object));
     }
@@ -1802,7 +1910,7 @@ void process_user_queue_item(snac *snac, xs_dict *q_item)
         /* iterate the recipients */
         p = rcpts;
         while (xs_list_iter(&p, &actor)) {
-            xs *inbox = get_actor_inbox(snac, actor);
+            xs *inbox = get_actor_inbox(actor);
 
             if (inbox != NULL) {
                 /* add to the set and, if it's not there, send message */
@@ -1879,10 +1987,8 @@ int process_user_queue(snac *snac)
     while (xs_list_iter(&p, &fn)) {
         xs *q_item = dequeue(fn);
 
-        if (q_item == NULL) {
-            snac_log(snac, xs_fmt("process_user_queue q_item error"));
+        if (q_item == NULL)
             continue;
-        }
 
         process_user_queue_item(snac, q_item);
         cnt++;
@@ -2018,6 +2124,71 @@ void process_queue_item(xs_dict *q_item)
 
         srv_log(xs_dup("purge end"));
     }
+    else
+    if (strcmp(type, "input") == 0) {
+        xs_dict *msg = xs_dict_get(q_item, "message");
+        xs_dict *req = xs_dict_get(q_item, "req");
+        int retries  = xs_number_get(xs_dict_get(q_item, "retries"));
+
+        /* do some instance-level checks */
+        int r = process_input_message(NULL, msg, req);
+
+        if (r == 0) {
+            /* transient error? retry */
+            int queue_retry_max = xs_number_get(xs_dict_get(srv_config, "queue_retry_max"));
+
+            if (retries > queue_retry_max)
+                srv_log(xs_fmt("shared input giving up"));
+            else {
+                /* reenqueue */
+                enqueue_shared_input(msg, req, retries + 1);
+                srv_log(xs_fmt("shared input requeue #%d", retries + 1));
+            }
+        }
+        else
+        if (r == 2) {
+            /* redistribute the input message to all users */
+            char *ntid = xs_dict_get(q_item, "ntid");
+            xs *tmpfn  = xs_fmt("%s/tmp/%s.json", srv_basedir, ntid);
+            FILE *f;
+
+            if ((f = fopen(tmpfn, "w")) != NULL) {
+                xs_json_dump(q_item, 4, f);
+                fclose(f);
+            }
+
+            xs *users = user_list();
+            xs_list *p = users;
+            char *v;
+            int cnt = 0;
+
+            while (xs_list_iter(&p, &v)) {
+                snac user;
+
+                if (user_open(&user, v)) {
+                    if (is_msg_for_me(&user, msg)) {
+                        xs *fn = xs_fmt("%s/queue/%s.json", user.basedir, ntid);
+
+                        snac_debug(&user, 1, xs_fmt("enqueue_input (from shared inbox) %s", fn));
+
+                        if (link(tmpfn, fn) < 0)
+                            srv_log(xs_fmt("link(%s, %s) error", tmpfn, fn));
+
+                        cnt++;
+                    }
+
+                    user_free(&user);
+                }
+            }
+
+            unlink(tmpfn);
+
+            if (cnt == 0) {
+                srv_archive_qitem("no_valid_recipients", q_item);
+                srv_debug(1, xs_fmt("no valid recipients for %s", tmpfn));
+            }
+        }
+    }
     else
         srv_log(xs_fmt("unexpected q_item type '%s'", type));
 }
@@ -2162,6 +2333,12 @@ int activitypub_post_handler(const xs_dict *req, const char *q_path,
         return 400;
     }
 
+    if (xs_is_null(payload)) {
+        *body  = xs_str_new("no payload");
+        *ctype = "text/plain";
+        return 400;
+    }
+
     if (xs_str_in(i_ctype, "application/activity+json") == -1 &&
         xs_str_in(i_ctype, "application/ld+json") == -1)
         return 0;
@@ -2191,6 +2368,11 @@ int activitypub_post_handler(const xs_dict *req, const char *q_path,
     /* get the user and path */
     xs *l = xs_split_n(q_path, "/", 2);
 
+    if (xs_list_len(l) == 2 && strcmp(xs_list_get(l, 1), "shared-inbox") == 0) {
+        enqueue_shared_input(msg, req, 0);
+        return 202;
+    }
+
     if (xs_list_len(l) != 3 || strcmp(xs_list_get(l, 2), "inbox") != 0) {
         /* strange q_path */
         srv_debug(1, xs_fmt("activitypub_post_handler unsupported path %s", q_path));
