@@ -1,5 +1,5 @@
 /* snac - A simple, minimalistic ActivityPub instance */
-/* copyright (c) 2022 - 2023 grunfink et al. / MIT license */
+/* copyright (c) 2022 - 2024 grunfink et al. / MIT license */
 
 #include "xs.h"
 #include "xs_hex.h"
@@ -99,6 +99,9 @@ int srv_open(char *basedir, int auto_upgrade)
     xs *ibdir = xs_fmt("%s/inbox", srv_basedir);
     mkdirx(ibdir);
 
+    xs *tmpdir = xs_fmt("%s/tmp", srv_basedir);
+    mkdirx(tmpdir);
+
 #ifdef __OpenBSD__
     char *v = xs_dict_get(srv_config, "disable_openbsd_security");
 
@@ -147,12 +150,12 @@ void user_free(snac *snac)
 }
 
 
-int user_open(snac *snac, const char *uid)
+int user_open(snac *user, const char *uid)
 /* opens a user */
 {
     int ret = 0;
 
-    memset(snac, '\0', sizeof(struct _snac));
+    *user = (snac){0};
 
     if (validate_uid(uid)) {
         xs *cfg_file = NULL;
@@ -171,52 +174,52 @@ int user_open(snac *snac, const char *uid)
                 xs *v2 = xs_tolower_i(xs_dup(v));
 
                 if (strcmp(lcuid, v2) == 0) {
-                    snac->uid = xs_dup(v);
+                    user->uid = xs_dup(v);
                     break;
                 }
             }
         }
         else
-            snac->uid = xs_str_new(uid);
+            user->uid = xs_str_new(uid);
 
-        if (snac->uid == NULL)
+        if (user->uid == NULL)
             return ret;
 
-        snac->basedir = xs_fmt("%s/user/%s", srv_basedir, snac->uid);
+        user->basedir = xs_fmt("%s/user/%s", srv_basedir, user->uid);
 
-        cfg_file = xs_fmt("%s/user.json", snac->basedir);
+        cfg_file = xs_fmt("%s/user.json", user->basedir);
 
         if ((f = fopen(cfg_file, "r")) != NULL) {
             /* read full config file */
-            snac->config = xs_json_load(f);
+            user->config = xs_json_load(f);
             fclose(f);
 
-            if (snac->config != NULL) {
-                xs *key_file = xs_fmt("%s/key.json", snac->basedir);
+            if (user->config != NULL) {
+                xs *key_file = xs_fmt("%s/key.json", user->basedir);
 
                 if ((f = fopen(key_file, "r")) != NULL) {
-                    snac->key = xs_json_load(f);
+                    user->key = xs_json_load(f);
                     fclose(f);
 
-                    if (snac->key != NULL) {
-                        snac->actor = xs_fmt("%s/%s", srv_baseurl, snac->uid);
-                        snac->md5   = xs_md5_hex(snac->actor, strlen(snac->actor));
+                    if (user->key != NULL) {
+                        user->actor = xs_fmt("%s/%s", srv_baseurl, user->uid);
+                        user->md5   = xs_md5_hex(user->actor, strlen(user->actor));
 
                         /* everything is ok right now */
                         ret = 1;
 
                         /* does it have a configuration override? */
-                        xs *cfg_file_o = xs_fmt("%s/user_o.json", snac->basedir);
+                        xs *cfg_file_o = xs_fmt("%s/user_o.json", user->basedir);
                         if ((f = fopen(cfg_file_o, "r")) != NULL) {
-                            snac->config_o = xs_json_load(f);
+                            user->config_o = xs_json_load(f);
                             fclose(f);
 
-                            if (snac->config_o == NULL)
+                            if (user->config_o == NULL)
                                 srv_log(xs_fmt("error parsing '%s'", cfg_file_o));
                         }
 
-                        if (snac->config_o == NULL)
-                            snac->config_o = xs_dict_new();
+                        if (user->config_o == NULL)
+                            user->config_o = xs_dict_new();
                     }
                     else
                         srv_log(xs_fmt("error parsing '%s'", key_file));
@@ -234,7 +237,7 @@ int user_open(snac *snac, const char *uid)
         srv_debug(1, xs_fmt("invalid user '%s'", uid));
 
     if (!ret)
-        user_free(snac);
+        user_free(user);
 
     return ret;
 }
@@ -646,10 +649,14 @@ int _object_add(const char *id, const xs_dict *obj, int ow)
     xs *fn     = _object_fn(id);
     FILE *f;
 
-    if (!ow && mtime(fn) > 0.0) {
-        /* object already here */
-        srv_debug(1, xs_fmt("object_add object already here %s", id));
-        return 204; /* No content */
+    if (mtime(fn) > 0.0) {
+        if (!ow) {
+            /* object already here */
+            srv_debug(1, xs_fmt("object_add object already here %s", id));
+            return 204; /* No content */
+        }
+        else
+            status = 200;
     }
 
     if ((f = fopen(fn, "w")) != NULL) {
@@ -1586,9 +1593,12 @@ void tag_index(const char *id, const xs_dict *obj)
             char *name = xs_dict_get(v, "name");
 
             if (!xs_is_null(type) && !xs_is_null(name) && strcmp(type, "Hashtag") == 0) {
-                if (*name == '#')
+                while (*name == '#' || *name == '@')
                     name++;
 
+                if (*name == '\0')
+                    continue;
+
                 name = xs_tolower_i(name);
 
                 xs *md5_tag   = xs_md5_hex(name, strlen(name));
@@ -1832,6 +1842,10 @@ void lastlog_write(snac *snac, const char *source)
 void inbox_add(const char *inbox)
 /* collects a shared inbox */
 {
+    /* don't collect ourselves */
+    if (xs_startswith(inbox, srv_baseurl))
+        return;
+
     xs *md5 = xs_md5_hex(inbox, strlen(inbox));
     xs *fn  = xs_fmt("%s/inbox/%s", srv_basedir, md5);
     FILE *f;
@@ -2123,6 +2137,21 @@ void enqueue_input(snac *snac, const xs_dict *msg, const xs_dict *req, int retri
 }
 
 
+void enqueue_shared_input(const xs_dict *msg, const xs_dict *req, int retries)
+/* enqueues an input message from the shared input */
+{
+    xs *qmsg   = _new_qmsg("input", msg, retries);
+    char *ntid = xs_dict_get(qmsg, "ntid");
+    xs *fn     = xs_fmt("%s/queue/%s.json", srv_basedir, ntid);
+
+    qmsg = xs_dict_append(qmsg, "req", req);
+
+    qmsg = _enqueue_put(fn, qmsg);
+
+    srv_debug(1, xs_fmt("enqueue_shared_input %s", fn));
+}
+
+
 void enqueue_output_raw(const char *keyid, const char *seckey,
                         xs_dict *msg, xs_str *inbox, int retries, int p_status)
 /* enqueues an output message to an inbox */
@@ -2139,7 +2168,7 @@ void enqueue_output_raw(const char *keyid, const char *seckey,
     qmsg = xs_dict_append(qmsg, "seckey", seckey);
 
     /* if it's to be sent right now, bypass the disk queue and post the job */
-    if (retries == 0 && job_fifo_ready())
+    if (retries == 0)
         job_post(qmsg, 0);
     else {
         qmsg = _enqueue_put(fn, qmsg);
@@ -2165,7 +2194,7 @@ void enqueue_output(snac *snac, xs_dict *msg, xs_str *inbox, int retries, int p_
 void enqueue_output_by_actor(snac *snac, xs_dict *msg, const xs_str *actor, int retries)
 /* enqueues an output message for an actor */
 {
-    xs *inbox = get_actor_inbox(snac, actor);
+    xs *inbox = get_actor_inbox(actor);
 
     if (!xs_is_null(inbox))
         enqueue_output(snac, msg, inbox, retries, 0);
@@ -2199,7 +2228,7 @@ void enqueue_telegram(const xs_str *msg, const char *bot, const char *chat_id)
 
     qmsg = _enqueue_put(fn, qmsg);
 
-    srv_debug(1, xs_fmt("enqueue_email %s %s", bot, chat_id));
+    srv_debug(1, xs_fmt("enqueue_telegram %s %s", bot, chat_id));
 }
 
 
@@ -2263,7 +2292,7 @@ void enqueue_request_replies(snac *user, const char *id)
 
     qmsg = _enqueue_put(fn, qmsg);
 
-    snac_debug(user, 1, xs_fmt("enqueue_request_replies %s", id));
+    snac_debug(user, 2, xs_fmt("enqueue_request_replies %s", id));
 }
 
 
@@ -2370,8 +2399,7 @@ xs_dict *dequeue(const char *fn)
 {
     xs_dict *obj = queue_get(fn);
 
-    if (obj != NULL)
-        unlink(fn);
+    unlink(fn);
 
     return obj;
 }
@@ -2516,6 +2544,14 @@ void purge_server(void)
             tag_gc += index_gc(v2);
             xs *bak = xs_fmt("%s.bak", v2);
             unlink(bak);
+
+            if (index_len(v2) == 0) {
+                /* there are no longer any entry with this tag;
+                   purge it completely */
+                unlink(v2);
+                xs *dottag = xs_replace(v2, ".idx", ".tag");
+                unlink(dottag);
+            }
         }
     }
 
@@ -2719,3 +2755,17 @@ void srv_archive_error(const char *prefix, const xs_str *err,
         fclose(f);
     }
 }
+
+
+void srv_archive_qitem(char *prefix, xs_dict *q_item)
+/* archives a q_item in the error folder */
+{
+    xs *ntid = tid(0);
+    xs *fn   = xs_fmt("%s/error/%s_qitem_%s", srv_basedir, ntid, prefix);
+    FILE *f;
+
+    if ((f = fopen(fn, "w")) != NULL) {
+        xs_json_dump(q_item, 4, f);
+        fclose(f);
+    }
+}
