@@ -1,5 +1,5 @@
 /* snac - A simple, minimalistic ActivityPub instance */
-/* copyright (c) 2022 - 2023 grunfink et al. / MIT license */
+/* copyright (c) 2022 - 2024 grunfink et al. / MIT license */
 
 #include "xs.h"
 #include "xs_io.h"
@@ -12,7 +12,7 @@
 int usage(void)
 {
     printf("snac " VERSION " - A simple, minimalistic ActivityPub instance\n");
-    printf("Copyright (c) 2022 - 2023 grunfink et al. / MIT license\n");
+    printf("Copyright (c) 2022 - 2024 grunfink et al. / MIT license\n");
     printf("\n");
     printf("Commands:\n");
     printf("\n");
@@ -27,7 +27,7 @@ int usage(void)
     printf("follow {basedir} {uid} {actor}      Follows an actor\n");
     printf("unfollow {basedir} {uid} {actor}    Unfollows an actor\n");
     printf("request {basedir} {uid} {url}       Requests an object\n");
-    printf("actor {basedir} {uid} {url}         Requests an actor\n");
+    printf("actor {basedir} [{uid}] {url}       Requests an actor\n");
     printf("note {basedir} {uid} {'text'}       Sends a note to followers\n");
     printf("resetpwd {basedir} {uid}            Resets the password of a user\n");
     printf("ping {basedir} {uid} {actor}        Pings an actor\n");
@@ -56,6 +56,16 @@ char *get_argv(int *argi, int argc, char *argv[])
 
 #define GET_ARGV() get_argv(&argi, argc, argv)
 
+#include "xs_html.h"
+
+xs_html *html_note(snac *user, char *summary,
+                   char *div_id, char *form_id,
+                   char *ta_plh, char *ta_content,
+                   char *edit_id, char *actor_id,
+                   xs_val *cw_yn, char *cw_text,
+                   xs_val *mnt_only, char *redir,
+                   char *in_reply_to, int poll);
+
 int main(int argc, char *argv[])
 {
     char *cmd;
@@ -169,6 +179,22 @@ int main(int argc, char *argv[])
         return 0;
     }
 
+    if (argi == argc && strcmp(cmd, "actor") == 0) { /** **/
+        /* query an actor without user (non-signed) */
+        xs *actor = NULL;
+        int status;
+
+        status = actor_request(NULL, user, &actor);
+
+        printf("status: %d\n", status);
+        if (valid_status(status)) {
+            xs_json_dump(actor, 4, stdout);
+            printf("\n");
+        }
+
+        return 0;
+    }
+
     if (!user_open(&snac, user)) {
         printf("invalid user '%s'\n", user);
         return 1;
@@ -184,6 +210,13 @@ int main(int argc, char *argv[])
         return deluser(&snac);
     }
 
+    if (strcmp(cmd, "update") == 0) { /** **/
+        xs *a_msg = msg_actor(&snac);
+        xs *u_msg = msg_update(&snac, a_msg);
+        enqueue_message(&snac, u_msg);
+        return 0;
+    }
+
     if (strcmp(cmd, "queue") == 0) { /** **/
         process_user_queue(&snac);
         return 0;
