@@ -1,5 +1,5 @@
 /* snac - A simple, minimalistic ActivityPub instance */
-/* copyright (c) 2022 - 2023 grunfink et al. / MIT license */
+/* copyright (c) 2022 - 2024 grunfink et al. / MIT license */
 
 #ifndef NO_MASTODON_API
 
@@ -279,8 +279,14 @@ int oauth_post_handler(const xs_dict *req, const char *q_path,
                     xs *code = random_str();
 
                     xs_free(*body);
-                    *body = xs_fmt("%s?code=%s", redir, code);
-                    status = 303;
+
+                    if (strcmp(redir, "urn:ietf:wg:oauth:2.0:oob") == 0) {
+                        *body = xs_dup(code);
+                    }
+                    else {
+                        *body = xs_fmt("%s?code=%s", redir, code);
+                        status = 303;
+                    }
 
                     /* if there is a state, add it */
                     if (!xs_is_null(state) && *state) {
@@ -592,10 +598,48 @@ xs_dict *mastoapi_account(const xs_dict *actor)
         acct = xs_dict_append(acct, "emojis", eml);
     }
 
+    acct = xs_dict_append(acct, "locked", xs_stock_false);
+    acct = xs_dict_append(acct, "followers_count", xs_stock_0);
+    acct = xs_dict_append(acct, "following_count", xs_stock_0);
+    acct = xs_dict_append(acct, "statuses_count", xs_stock_0);
+
+    xs *fields = xs_list_new();
+    p = xs_dict_get(actor, "attachment");
+    xs_dict *v;
+
+    while (xs_list_iter(&p, &v)) {
+        char *type  = xs_dict_get(v, "type");
+        char *name  = xs_dict_get(v, "name");
+        char *value = xs_dict_get(v, "value");
+
+        if (!xs_is_null(type) && !xs_is_null(name) &&
+            !xs_is_null(value) && strcmp(type, "PropertyValue") == 0) {
+            xs *d = xs_dict_new();
+
+            d = xs_dict_append(d, "name", name);
+            d = xs_dict_append(d, "value", value);
+            d = xs_dict_append(d, "verified_at", xs_stock_null);
+
+            fields = xs_list_append(fields, d);
+        }
+    }
+
+    acct = xs_dict_append(acct, "fields", fields);
+
     return acct;
 }
 
 
+xs_str *mastoapi_date(char *date)
+/* converts an ISO 8601 date to whatever format Mastodon uses */
+{
+    xs_str *s = xs_crop_i(xs_dup(date), 0, 19);
+    s = xs_str_cat(s, ".000Z");
+
+    return s;
+}
+
+
 xs_dict *mastoapi_poll(snac *snac, const xs_dict *msg)
 /* creates a mastoapi Poll object */
 {
@@ -607,7 +651,8 @@ xs_dict *mastoapi_poll(snac *snac, const xs_dict *msg)
     xs *options = xs_list_new();
 
     poll = xs_dict_append(poll, "id", mid);
-    poll = xs_dict_append(poll, "expires_at", xs_dict_get(msg, "endTime"));
+    xs *fd = mastoapi_date(xs_dict_get(msg, "endTime"));
+    poll = xs_dict_append(poll, "expires_at", fd);
     poll = xs_dict_append(poll, "expired",
             xs_dict_get(msg, "closed") != NULL ? xs_stock_true : xs_stock_false);
 
@@ -673,9 +718,11 @@ xs_dict *mastoapi_status(snac *snac, const xs_dict *msg)
     st = xs_dict_append(st, "id",           mid);
     st = xs_dict_append(st, "uri",          id);
     st = xs_dict_append(st, "url",          id);
-    st = xs_dict_append(st, "created_at",   xs_dict_get(msg, "published"));
     st = xs_dict_append(st, "account",      acct);
 
+    xs *fd = mastoapi_date(xs_dict_get(msg, "published"));
+    st = xs_dict_append(st, "created_at", fd);
+
     {
         const char *content = xs_dict_get(msg, "content");
         const char *name    = xs_dict_get(msg, "name");
@@ -928,10 +975,15 @@ xs_dict *mastoapi_status(snac *snac, const xs_dict *msg)
     st = xs_dict_append(st, "text", tmp);
 
     tmp = xs_dict_get(msg, "updated");
+    xs *fd2 = NULL;
     if (xs_is_null(tmp))
         tmp = xs_stock_null;
+    else {
+        fd2 = mastoapi_date(tmp);
+        tmp = fd2;
+    }
 
-    st = xs_dict_append(st, "edited_at", tmp);
+    st = xs_dict_append(st, "edited_at", fd2);
 
     if (strcmp(type, "Question") == 0) {
         xs *poll = mastoapi_poll(snac, msg);
@@ -1104,6 +1156,10 @@ int mastoapi_get_handler(const xs_dict *req, const char *q_path,
                 acct = xs_dict_set(acct, "fields", fields);
             }
 
+            acct = xs_dict_append(acct, "followers_count", xs_stock_0);
+            acct = xs_dict_append(acct, "following_count", xs_stock_0);
+            acct = xs_dict_append(acct, "statuses_count", xs_stock_0);
+
             *body  = xs_json_dumps(acct, 4);
             *ctype = "application/json";
             status = 200;
@@ -1368,12 +1424,16 @@ int mastoapi_get_handler(const xs_dict *req, const char *q_path,
                 if (!xs_match(type, "Note|Question|Page|Article"))
                     continue;
 
-                const char *from;
+                const char *from = NULL;
                 if (strcmp(type, "Page") == 0)
                     from = xs_dict_get(msg, "audience");
-                else
+
+                if (from == NULL)
                     from = xs_dict_get(msg, "attributedTo");
 
+                if (from == NULL)
+                    continue;
+
                 /* is this message from a person we don't follow? */
                 if (strcmp(from, snac1.actor) && !following_check(&snac1, from)) {
                     /* discard if it was not boosted */
