@@ -1,7 +1,7 @@
 /* snac - A simple, minimalistic ActivityPub instance */
-/* copyright (c) 2022 - 2023 grunfink et al. / MIT license */
+/* copyright (c) 2022 - 2024 grunfink et al. / MIT license */
 
-#define VERSION "2.43"
+#define VERSION "2.44"
 
 #define USER_AGENT "snac/" VERSION
 
@@ -12,6 +12,10 @@
 
 #define ISO_DATE_SPEC "%Y-%m-%dT%H:%M:%SZ"
 
+#ifndef MAX_THREADS
+#define MAX_THREADS 256
+#endif
+
 extern double disk_layout;
 extern xs_str *srv_basedir;
 extern xs_dict *srv_config;
@@ -31,7 +35,7 @@ void srv_log(xs_str *str);
 #define srv_debug(level, str) do { if (dbglevel >= (level)) \
     { srv_log((str)); } } while (0)
 
-typedef struct _snac {
+typedef struct {
     xs_str *uid;        /* uid */
     xs_str *basedir;    /* user base directory */
     xs_dict *config;    /* user configuration */
@@ -41,6 +45,16 @@ typedef struct _snac {
     xs_str *md5;        /* actor url md5 */
 } snac;
 
+typedef struct {
+    int srv_running;        /* server running on/off */
+    int use_fcgi;           /* FastCGI use on/off */
+    time_t srv_start_time;  /* start time */
+    int job_fifo_size;      /* job fifo size */
+    int peak_job_fifo_size; /* maximum job fifo size seen */
+    int n_threads;          /* number of configured threads */
+    enum { THST_WAIT, THST_IN, THST_QUEUE, THST_STOP } th_state[MAX_THREADS];
+} srv_state;
+
 void snac_log(snac *user, xs_str *str);
 #define snac_debug(user, level, str) do { if (dbglevel >= (level)) \
     { snac_log((user), (str)); } } while (0)
@@ -64,6 +78,7 @@ void srv_archive(const char *direction, const char *url, xs_dict *req,
                  const char *body, int b_size);
 void srv_archive_error(const char *prefix, const xs_str *err,
                        const xs_dict *req, const xs_val *data);
+void srv_archive_qitem(char *prefix, xs_dict *q_item);
 
 double mtime_nl(const char *fn, int *n_link);
 #define mtime(fn) mtime_nl(fn, NULL)
@@ -181,6 +196,7 @@ int instance_block(const char *instance);
 int instance_unblock(const char *instance);
 
 void enqueue_input(snac *snac, const xs_dict *msg, const xs_dict *req, int retries);
+void enqueue_shared_input(const xs_dict *msg, const xs_dict *req, int retries);
 void enqueue_output_raw(const char *keyid, const char *seckey,
                         xs_dict *msg, xs_str *inbox, int retries, int p_status);
 void enqueue_output(snac *snac, xs_dict *msg, xs_str *inbox, int retries, int p_status);
@@ -211,7 +227,7 @@ xs_dict *http_signed_request(snac *snac, const char *method, const char *url,
                             const char *body, int b_size,
                             int *status, xs_str **payload, int *p_size,
                             int timeout);
-int check_signature(snac *snac, xs_dict *req, xs_str **err);
+int check_signature(xs_dict *req, xs_str **err);
 
 void httpd(void);
 
@@ -241,14 +257,14 @@ xs_dict *msg_question(snac *user, const char *content, xs_list *attach,
                       const xs_list *opts, int multiple, int end_secs);
 
 int activitypub_request(snac *snac, const char *url, xs_dict **data);
-int actor_request(snac *snac, const char *actor, xs_dict **data);
+int actor_request(snac *user, const char *actor, xs_dict **data);
 void timeline_request_replies(snac *user, const char *id);
 int send_to_inbox_raw(const char *keyid, const char *seckey,
                   const xs_str *inbox, const xs_dict *msg,
                   xs_val **payload, int *p_size, int timeout);
 int send_to_inbox(snac *snac, const xs_str *inbox, const xs_dict *msg,
                   xs_val **payload, int *p_size, int timeout);
-xs_str *get_actor_inbox(snac *snac, const char *actor);
+xs_str *get_actor_inbox(const char *actor);
 int send_to_actor(snac *snac, const char *actor, const xs_dict *msg,
                   xs_val **payload, int *p_size, int timeout);
 int is_msg_public(const xs_dict *msg);
@@ -266,7 +282,6 @@ int activitypub_post_handler(const xs_dict *req, const char *q_path,
 
 xs_str *not_really_markdown(const char *content, xs_list **attach);
 xs_str *sanitize(const char *content);
-xs_str *encode_html_strict(const char *str);
 xs_str *encode_html(const char *str);
 
 xs_str *html_timeline(snac *user, const xs_list *list, int local,
@@ -285,7 +300,6 @@ int deluser(snac *user);
 
 extern const char *snac_blurb;
 
-int job_fifo_ready(void);
 void job_post(const xs_val *job, int urgent);
 void job_wait(xs_val **job);
 
